use std::collections::HashMap;

use crate::id_of;
use mongodb::{bson::doc, Collection};

use crate::variant::Variant;

// todo : api key auth

// ITEM
// VARIANTS
// QUANTIZATION
// IMPORT / EXPORT
// DEPENDENCIES
// DEMAND STATS
// SEASONAL REVIEWS

fn hashmap_to_bson_document(
    hashmap: HashMap<String, mongodb::bson::Document>,
) -> mongodb::bson::Document {
    let mut document = mongodb::bson::Document::new();
    for (key, value) in hashmap {
        document.insert(key, value);
    }
    document
}

pub struct ItemEntry {
    pub name: String,
    pub category: String,
    pub variants: HashMap<String, Variant>,
}

impl ItemEntry {
    pub fn new(doc: &mdq::Document) -> Self {
        let name = std::path::Path::new(&doc.path)
            .file_stem()
            .unwrap()
            .to_str()
            .unwrap()
            .to_string();
        let category = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("category")
            .unwrap()
            .as_str()
            .unwrap()
            .to_string();

        let mut variants = HashMap::new();

        for (variant_name, variant) in doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("variants")
            .map(|x| x.as_mapping().unwrap().clone())
            .unwrap_or_default()
        {
            variants.insert(
                variant_name.as_str().unwrap().to_string(),
                Variant::from_yml(&variant, variant_name.as_str().unwrap(), &name),
            );
        }

        Self {
            name,
            category,
            variants,
        }
    }

    pub async fn init_db(&self, mongodb: &mongodb::Client) {
        log::info!("Adding item {} to DB", self.name);
        let items: Collection<mongodb::bson::Document> =
            mongodb.database("cdb").collection("items");

        let doc = mongodb::bson::doc! {
            "_id": self.name.clone(),
            "name": self.name.clone(),
            "category": self.category.clone(),
            "variants": hashmap_to_bson_document(self.variants.iter().map(|x| (x.0.clone(), x.1.as_bson())).collect())
        };

        if items
            .find_one(id_of!(&self.name), None)
            .await
            .unwrap()
            .is_none()
        {
            items.insert_one(doc, None).await.unwrap();
        } else {
            items
                .find_one_and_update(id_of!(&self.name), doc! { "$set": doc }, None)
                .await
                .unwrap();
        }
    }
}

/// Represents a single item in a collection of physical goods.
///
/// This struct serves as a blueprint for describing individual items within
/// a larger inventory or database of physical goods. It includes fields for
/// the name of the item and its category.
pub struct Item {
    pub item: ItemEntry,
}

impl Item {
    pub const fn new(item: ItemEntry) -> Self {
        Self { item }
    }

    pub fn get_variants(&self) -> Vec<String> {
        self.item.variants.keys().cloned().collect()
    }

    pub fn variant(&self, variant: &str) -> Option<Variant> {
        self.item.variants.get(variant).cloned()
    }
}
